@@ -1,16 +1,12 @@
-import copy
 import itertools
-import json
 import logging
 import numpy as np
-import os
 import pandas as pd
-import tempfile
+from sklearn.neighbors._kd_tree import KDTree
 import time
-import uuid
 
-from datamart_materialize.d3m import d3m_metadata
 from datamart_materialize import types
+from datamart_profiler.spatial import median_smallest_distance
 from datamart_profiler.temporal import get_temporal_resolution, \
     temporal_aggregation_keys
 
@@ -23,6 +19,34 @@ class AugmentationError(ValueError):
     """
 
 
+class WriteCounter(object):
+    """File wrapper that counts the number of bytes written.
+    """
+    def __init__(self, inner):
+        self.inner = inner
+        self.size = 0
+
+    def __iter__(self):
+        # Pandas needs file objects to have __iter__
+        return self
+
+    def write(self, buf):
+        self.size += len(buf)
+        return self.inner.write(buf)
+
+    def flush(self):
+        return self.inner.flush()
+
+    def close(self):
+        self.inner.close()
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        return self.inner.__exit__(exc_type, exc_val, exc_tb)
+
+
 class _UniqueIndexKey(object):
     def __repr__(self):
         return "UNIQUE_INDEX_KEY"
@@ -40,11 +64,14 @@ temporal_resolutions_priorities = {
 def _transform_index(level, func):
     def inner(index):
         if isinstance(index, pd.MultiIndex):
-            return index.set_levels(
-                [index.levels[i] if i != level
-                 else func(index.levels[i])
+            old_index = index.to_frame()
+            return pd.MultiIndex.from_arrays(
+                [old_index.iloc[:, i] if i != level
+                 else func(old_index.iloc[:, i])
                  for i in range(len(index.levels))]
             )
+        elif isinstance(index, pd.Series):
+            return func(pd.Index(index))
         else:
             return func(index)
 
@@ -53,9 +80,10 @@ def _transform_index(level, func):
 
 def _transform_data_index(data, level, func):
     if isinstance(data.index, pd.MultiIndex):
-        data.index = data.index.set_levels(
-            [data.index.levels[i] if i != level
-             else func(data.index.levels[i])
+        old_index = data.index.to_frame()
+        data.index = pd.MultiIndex.from_arrays(
+            [old_index.iloc[:, i] if i != level
+             else func(old_index.iloc[:, i])
              for i in range(len(data.index.levels))]
         )
     else:
@@ -97,6 +125,9 @@ def set_data_index(data, columns, columns_metadata, drop=False):
                 lambda idx: idx.str.lower(),
             )
 
+    # Names of multiindex have to match for join() to work
+    data.index.names = ['%04d' % i for i in range(len(data.index.names))]
+
     return data
 
 
@@ -124,10 +155,14 @@ def match_temporal_resolutions(input_data, companion_data, temporal_resolution=N
             else:
                 funcs.append(lambda x: x)
 
-        return lambda idx: idx.set_levels(
-            func(lvl)
-            for lvl, func in zip(idx.levels, funcs)
-        )
+        def transform(index):
+            old_index = index.to_frame()
+            return pd.MultiIndex.from_arrays(
+                func(old_index.iloc[:, lvl])
+                for lvl, func in zip(range(len(index.levels)), funcs)
+            )
+
+        return transform
     elif (isinstance(input_data.index, pd.DatetimeIndex)
           and isinstance(companion_data.index, pd.DatetimeIndex)):
         return match_column_temporal_resolutions(
@@ -145,6 +180,12 @@ def match_column_temporal_resolutions(index_1, index_2, level,
     """Matches the resolutions between the dataset indices.
     """
 
+    col = ''
+    if isinstance(index_1, pd.MultiIndex):
+        index_1 = index_1.levels[level]
+        index_2 = index_2.levels[level]
+        col = " (level %d)" % level
+
     if not (index_1.is_all_dates and index_2.is_all_dates):
         return lambda idx: idx
 
@@ -166,7 +207,11 @@ def match_column_temporal_resolutions(index_1, index_2, level,
         if (temporal_resolutions_priorities[resolution_1] >
                 temporal_resolutions_priorities[resolution_2]):
             # Change resolution of second index to the first's
-            logger.info("Temporal alignment: right to '%s'", resolution_1)
+            logger.info(
+                "Temporal alignment: right to '%s'%s",
+                resolution_1,
+                col,
+            )
             key = temporal_aggregation_keys[resolution_1]
             if isinstance(key, str):
                 return _transform_index(level, lambda idx: idx.strftime(key))
@@ -174,7 +219,11 @@ def match_column_temporal_resolutions(index_1, index_2, level,
                 return _transform_index(level, lambda idx: idx.map(key))
         else:
             # Change resolution of first index to the second's
-            logger.info("Temporal alignment: left to '%s'", resolution_2)
+            logger.info(
+                "Temporal alignment: left to '%s'%s",
+                resolution_2,
+                col,
+            )
             key = temporal_aggregation_keys[resolution_2]
             if isinstance(key, str):
                 return _transform_index(level, lambda idx: idx.strftime(key))
@@ -290,32 +339,118 @@ def perform_aggregations(
 CHUNK_SIZE_ROWS = 10000
 
 
+def _tree_nearest(tree, max_dist):
+    def transform(df):
+        # Convert to numeric numpy array
+        points = pd.DataFrame({
+            'x': pd.to_numeric(
+                df.iloc[:, 0],
+                errors='coerce',
+                downcast='float',
+            ),
+            'y': pd.to_numeric(
+                df.iloc[:, 1],
+                errors='coerce',
+                downcast='float',
+            ),
+        }).values
+        # Run input through tree
+        dist, indices = tree.query(points, return_distance=True)
+        indices = indices.reshape((-1,))
+        dist = dist.reshape((-1,))
+
+        # Build array of transformed coordinates
+        coords = tree.get_arrays()[0]
+        res = coords[indices]
+
+        # Discard points too far
+        res[dist >= max_dist] = np.nan
+        return res
+
+    return transform
+
+
+KEEP_COLUMN_FIELDS = {'name', 'structural_type', 'semantic_types'}
+
+
 def join(
     original_data, augment_data_path, original_metadata, augment_metadata,
-    destination_csv,
+    writer,
     left_columns, right_columns,
     how='left', columns=None,
     agg_functions=None, temporal_resolution=None,
     return_only_datamart_data=False,
 ):
     """
-    Performs a join between original_data (pandas.DataFrame)
+    Performs a join between original_data (pandas.DataFrame or path to CSV)
     and augment_data (pandas.DataFrame) using left_columns and right_columns.
 
-    Returns the new pandas.DataFrame object.
+    The result is written to the writer object.
+
+    Returns the metadata for the result.
     """
 
+    if isinstance(original_data, pd.DataFrame):
+        pass
+    elif hasattr(original_data, 'read'):
+        original_data = pd.read_csv(
+            original_data,
+            error_bad_lines=False,
+            dtype=str,
+        )
+    else:
+        raise TypeError(
+            "join() argument 1 should be a file or a DataFrame, got "
+            "%r" % type(original_data)
+        )
+
     augment_data_columns = [col['name'] for col in augment_metadata['columns']]
 
     # only converting data types for columns involved in augmentation
     original_join_columns_idx = []
     augment_join_columns_idx = []
+    augment_columns_transform = []
     for left, right in zip(left_columns, right_columns):
-        if len(left) > 1 or len(right) > 1:
+        if len(left) == 2 and len(right) == 2:
+            # Spatial augmentation
+            # Get those columns
+            points = original_data.iloc[:, left]
+            # De-duplicate
+            points = pd.DataFrame(list(
+                set(tuple(p) for p in points.values)
+            ))
+            # Convert to numeric numpy array
+            points = pd.DataFrame({
+                'x': pd.to_numeric(
+                    points.iloc[:, 0],
+                    errors='coerce',
+                    downcast='float',
+                ),
+                'y': pd.to_numeric(
+                    points.iloc[:, 1],
+                    errors='coerce',
+                    downcast='float',
+                ),
+            }).values
+            # Build KDTree
+            tree = KDTree(points)
+            # Compute max distance for nearest join
+            max_dist = 2 * median_smallest_distance(points, tree)
+            logger.info("Using nearest spatial join, max=%r", max_dist)
+            # Store transformation
+            augment_columns_transform.append((
+                right,
+                _tree_nearest(tree, max_dist),
+            ))
+
+            original_join_columns_idx.extend(left)
+            augment_join_columns_idx.extend(right)
+        elif len(left) > 1 or len(right) > 1:
             raise AugmentationError("Datamart currently does not support "
                                     "combination of columns for augmentation.")
-        original_join_columns_idx.append(left[0])
-        augment_join_columns_idx.append(right[0])
+        else:
+            original_join_columns_idx.append(left[0])
+            augment_join_columns_idx.append(right[0])
 
     original_data = set_data_index(
         original_data,
@@ -364,9 +499,13 @@ def join(
     start = time.perf_counter()
     join_ = []
     # Iterate over chunks of augment data
-    for i, augment_data in enumerate(
-            itertools.chain([first_augment_data], augment_data_chunks)
+    for augment_data in itertools.chain(
+            [first_augment_data], augment_data_chunks
     ):
+        # Run transforms
+        for cols, transform in augment_columns_transform:
+            augment_data.iloc[:, cols] = transform(augment_data.iloc[:, cols])
+
         # Convert data types
         augment_data = set_data_index(
             augment_data,
@@ -464,92 +603,132 @@ def join(
             qualValueType='dict'
         ))
 
-    join_.to_csv(destination_csv, index=False)
+    with WriteCounter(writer.open_file('w')) as fout:
+        join_.to_csv(fout, index=False, line_terminator='\r\n')
+        size = fout.size
 
     # Build a dict of information about all columns
     columns_metadata = dict()
+    for column in augment_metadata['columns']:
+        for agg in itertools.chain(
+            (None,),
+            AGGREGATION_FUNCTIONS,
+        ):
+            for name in (column['name'], column['name'] + '_r'):
+                column_metadata = {
+                    k: v for k, v in column.items()
+                    if k in KEEP_COLUMN_FIELDS
+                }
+                if agg is not None:
+                    name = agg + ' ' + name
+                column_metadata['name'] = name
+                if agg in {'sum', 'mean'}:
+                    column_metadata['structural_type'] = types.FLOAT
+                    column_metadata['semantic_types'] = []
+                elif agg == 'count':
+                    column_metadata['structural_type'] = types.INTEGER
+                    column_metadata['semantic_types'] = []
+                columns_metadata[name] = column_metadata
     for column in original_metadata['columns']:
         columns_metadata[column['name']] = column
-    for column in augment_metadata['columns']:
-        names = [
-            column['name'],
-            column['name'] + '_r'
-        ]
-        # agg names
-        all_names = itertools.chain(names, (
-            agg + ' ' + name
-            for agg, name in itertools.product(AGGREGATION_FUNCTIONS, names)
-        ))
-        for name in all_names:
-            column_metadata = copy.deepcopy(column)
-            column_metadata['name'] = name
-            if ('sum' in name or 'mean' in name
-                    or 'max' in name or 'min' in name):
-                column_metadata['structural_type'] = types.FLOAT
-            columns_metadata[name] = column_metadata
 
     # Then construct column metadata by looking them up in the dict
     columns_metadata = [columns_metadata[name] for name in join_.columns]
 
     return {
         'columns': columns_metadata,
-        'size': os.path.getsize(destination_csv),
+        'size': size,
         'qualities': qualities_list,
     }
 
 
 def union(original_data, augment_data_path, original_metadata, augment_metadata,
-          destination_csv,
+          writer,
           left_columns, right_columns,
           return_only_datamart_data=False):
     """
-    Performs a union between original_data (pandas.DataFrame)
+    Performs a union between original_data (pandas.DataFrame or path to CSV)
     and augment_data_path (path to CSV file) using columns.
 
-    Returns the new pandas.DataFrame object.
+    The result is streamed to the writer object.
+
+    Returns the metadata for the result.
     """
 
+    if isinstance(original_data, pd.DataFrame):
+        original_data = iter((original_data,))
+    elif hasattr(original_data, 'read'):
+        original_data = iter(pd.read_csv(
+            original_data,
+            error_bad_lines=False,
+            dtype=str,
+            chunksize=CHUNK_SIZE_ROWS,
+        ))
+    else:
+        raise TypeError(
+            "union() argument 1 should be a file or a DataFrame, got "
+            "%r" % type(original_data)
+        )
+
+    first_original_data = next(original_data)
+
     augment_data_columns = [col['name'] for col in augment_metadata['columns']]
 
     logger.info(
         "Performing union, original_data: %r, augment_data: %r, "
         "left_columns: %r, right_columns: %r",
-        original_data.columns, augment_data_columns,
+        first_original_data.columns, augment_data_columns,
         left_columns, right_columns,
     )
 
     # Column renaming
     rename = dict()
     for left, right in zip(left_columns, right_columns):
-        rename[augment_data_columns[right[0]]] = original_data.columns[left[0]]
+        rename[augment_data_columns[right[0]]] = \
+            first_original_data.columns[left[0]]
 
     # Missing columns will be created as NaN
     missing_columns = list(
-        set(original_data.columns) - set(augment_data_columns)
+        set(first_original_data.columns) - set(augment_data_columns)
     )
 
     # Sequential d3mIndex if needed, picking up from the last value
     # FIXME: Generated d3mIndex might collide with other splits?
     d3m_index = None
-    if 'd3mIndex' in original_data.columns:
-        d3m_index = int(original_data['d3mIndex'].max() + 1)
+    if 'd3mIndex' in first_original_data.columns:
+        d3m_index = int(first_original_data['d3mIndex'].max()) + 1
 
     logger.info("renaming: %r, missing_columns: %r", rename, missing_columns)
 
     # Streaming union
     start = time.perf_counter()
-    with open(destination_csv, 'w', newline='') as fout:
+    with WriteCounter(writer.open_file('w')) as fout:
+        orig_rows = 0
+        # Write header
+        fout.write(','.join(first_original_data.columns) + '\n')
         # Write original data
-        fout.write(','.join(original_data.columns) + '\n')
-        total_rows = 0
         if not return_only_datamart_data:
-            original_data.to_csv(fout, index=False, header=False)
-            total_rows += len(original_data)
+            for chunk in itertools.chain([first_original_data], original_data):
+                chunk.to_csv(
+                    fout,
+                    header=False,
+                    index=False,
+                    line_terminator='\r\n',
+                )
+                orig_rows += len(chunk)
+                if d3m_index is not None:
+                    d3m_index = max(
+                        d3m_index,
+                        int(chunk['d3mIndex'].max()) + 1,
+                    )
+
+        total_rows = orig_rows
 
         # Iterate on chunks of augment data
         augment_data_chunks = pd.read_csv(
             augment_data_path,
             error_bad_lines=False,
+            dtype=str,
             chunksize=CHUNK_SIZE_ROWS,
         )
         for augment_data in augment_data_chunks:
@@ -569,95 +748,35 @@ def union(original_data, augment_data_path, original_metadata, augment_metadata,
                 augment_data[name] = np.nan
 
             # Reorder columns
-            augment_data = augment_data[original_data.columns]
+            augment_data = augment_data[first_original_data.columns]
 
             # Add to CSV output
-            augment_data.to_csv(fout, index=False, header=False)
+            augment_data.to_csv(
+                fout,
+                header=False,
+                index=False,
+                line_terminator='\r\n',
+            )
             total_rows += len(augment_data)
+
+        size = fout.size
     logger.info("Union completed in %.4fs", time.perf_counter() - start)
 
     return {
-        'columns': original_metadata['columns'],
-        'size': os.path.getsize(destination_csv),
+        'columns': [
+            {k: v for k, v in col.items() if k in KEEP_COLUMN_FIELDS}
+            for col in original_metadata['columns']
+        ],
+        'size': size,
         'qualities': [dict(
             qualName='augmentation_info',
             qualValue=dict(
                 new_columns=[],
                 removed_columns=[],
-                nb_rows_before=original_data.shape[0],
+                nb_rows_before=orig_rows,
                 nb_rows_after=total_rows,
                 augmentation_type='union'
             ),
             qualValueType='dict'
         )],
     }
-
-
-def augment(data, newdata, metadata, task, columns=None, destination=None,
-            return_only_datamart_data=False):
-    """
-    Augments original data based on the task.
-
-    :param data: the data to be augmented, as binary file object.
-    :param newdata: the path to the CSV file to augment with.
-    :param metadata: the metadata of the data to be augmented.
-    :param task: the augmentation task.
-    :param columns: a list of column indices from newdata that will be added to data
-    :param destination: location to save the files.
-    :param return_only_datamart_data: only returns the portion of newdata that matches
-      well with data.
-    """
-
-    if 'id' not in task:
-        raise AugmentationError("Dataset id for the augmentation task not provided")
-
-    # TODO: add support for combining multiple columns before an augmentation
-    #   e.g.: [['street number', 'street', 'city']] and [['address']]
-    #   currently, Datamart does not support such cases
-    #   this means that spatial joins (with GPS) are not supported for now
-
-    # Prepare output D3M structure
-    if destination is None:
-        destination = tempfile.mkdtemp(prefix='datamart_aug_')
-    os.mkdir(destination)
-    os.mkdir(os.path.join(destination, 'tables'))
-    destination_csv = os.path.join(destination, 'tables', 'learningData.csv')
-    destination_metadata = os.path.join(destination, 'datasetDoc.json')
-
-    # Perform augmentation
-    start = time.perf_counter()
-    if task['augmentation']['type'] == 'join':
-        output_metadata = join(
-            pd.read_csv(data, error_bad_lines=False),
-            newdata,
-            metadata,
-            task['metadata'],
-            destination_csv,
-            task['augmentation']['left_columns'],
-            task['augmentation']['right_columns'],
-            columns=columns,
-            agg_functions=task['augmentation'].get('agg_functions'),
-            temporal_resolution=task['augmentation'].get('temporal_resolution'),  # look
-            return_only_datamart_data=return_only_datamart_data,
-        )
-    elif task['augmentation']['type'] == 'union':
-        output_metadata = union(
-            pd.read_csv(data, error_bad_lines=False),
-            newdata,
-            metadata,
-            task['metadata'],
-            destination_csv,
-            task['augmentation']['left_columns'],
-            task['augmentation']['right_columns'],
-            return_only_datamart_data=return_only_datamart_data,
-        )
-    else:
-        raise AugmentationError("Augmentation task not provided")
-    logger.info("Total augmentation: %.4fs", time.perf_counter() - start)
-
-    # Write out the D3M metadata
-    d3m_meta = d3m_metadata(uuid.uuid4().hex, output_metadata)
-    with open(destination_metadata, 'w') as fp:
-        json.dump(d3m_meta, fp, sort_keys=True, indent=2)
-
-    return destination
